import logging

from .database import BackupDB
from .sourcefile import SourceFile

log = logging.getLogger(__name__)

class Verifier:
    def __init__(self, db:BackupDB):
        self.db:BackupDB = db

    def verify(self):
        self.db.require_disk()
        self.db.current_disk.migrate_data()

        bad_objects = []
        try:
            for hash, fpath, stat in self.db.current_disk.walk_data():
                row = self.db.query_one('SELECT last_modified, size, virtual_path, ino FROM file_tree WHERE hash = ? LIMIT 1', (hash,))
                if not row:
                    log.warning(f'{hash} deleted or not in database')
                    continue

                lastmod, size, path, ino = row
                try:
                    srcf = SourceFile(fpath, fpath.lstat())
                except OSError as e:
                    log.warning(f'cannot read {fpath} ({e})')
                    continue

                if srcf.size != size:
                    msg = f'ERROR: {self.db.current_disk.name}/{path}: size mismatch ({srcf.size} != {size})'
                    log.error(msg)
                    bad_objects.append((path, hash))
                    continue

                log.info(f'check {path} ({hash})')
                srcf.copy()
                if srcf.hash != hash:
                    msg = f'ERROR: {self.db.current_disk.name}/{path}: hash ({srcf.hash} != {hash})'
                    log.error(msg)
                    bad_objects.append((path, hash))
        except KeyboardInterrupt:
            pass

        if bad_objects:
            log.info('')
            log.info('Bad files:')
            for path, hash in bad_objects:
                log.info(f'  {hash} {path}')
        else:
            log.info('No bad files.')



import time
import logging

from typing import Optional
from dataclasses import dataclass, field
from ktpanda.textcolor import Colorizer

from .utils import addcomma, disk_in_nexus, round_blocks, prettysize, prettysize_clr, size_color, Progress
from .nexus import NexusLevelDict
from .database import BackupDB

log = logging.getLogger(__name__)

@dataclass
class FileNode:
    '''Track the size of a file or folder for `ncdu` and `tree` commands'''

    name: str
    parent: object
    size: int = 0
    hash: Optional[str] = None
    lastmod: Optional[float] = None
    iscanon: bool = True
    refs: Optional[int] = None
    copies: Optional[int] = None
    nexus: Optional[str] = None
    priority: int = 1
    obj_priority: int = 1
    maxcopies: Optional[int] = None
    obj_maxcopies: Optional[int] = None
    inode: Optional[int]  = None
    children: dict[str, object] = field(default_factory=dict)

    total_files: int = 0
    total_size: int = 0
    copy_files: Optional[list] = None
    copy_size: Optional[list] = None

class Reporter:
    def __init__(self, db:BackupDB, ignore_lock=False):
        self.db:BackupDB = db

    def report(self, nexus_levels:NexusLevelDict=None):

        if nexus_levels is None:
            nexus_levels = NexusLevelDict(self.db.load_nexus())

        copy_data = nexus_levels.get_copy_data()
        if not copy_data or copy_data == [0]:
            yield Colorizer().t('No files in database.')
            return

        yield Colorizer().t('Usage:')

        for i, size in enumerate(copy_data):
            yield Colorizer().t(f'{i:3d}: ').t(prettysize_clr(size))

        clridx = 0
        yield Colorizer().t('  %-12s  %-20s %-7s %s' % ('name', 'total', 'free',' '.join('%-7d' % j for j in range(1, len(copy_data)))))
        with self.db.execute('SELECT name, nexus_index, size FROM disk WHERE name NOTNULL ORDER BY name') as curs:
            for name, index, dsize in curs:
                data = nexus_levels.get_copy_data(index)[1:]
                total = sum(data)
                free = dsize - total

                fg_clr = '!ccc' if (clridx & 1) else '!fff'
                bg_clr = '222' if (clridx & 1) else '333'
                clridx += 1

                txt = Colorizer(fg=fg_clr, bg=bg_clr)
                if self.db.current_disk and name == self.db.current_disk.name:
                    txt.set(fg='!c0f').t('> ')
                else:
                    txt.t('  ')

                txt.t(name, rpad=12).t(' ')
                txt.t(addcomma(total), lpad=20, fg=size_color(total)).t(' ')
                txt.t(prettysize_clr(free), lpad=7)
                if not data:
                    txt.t(' no files')
                else:
                    for size in data:
                        txt.t(' ').t(prettysize_clr(size), lpad=7)

                yield txt

    def print_report(self, *a, indent='', **kw):
        for line in self.report(*a, **kw):
            log.info(indent + line.get())

    def list_files(self, pat):
        destid = self.db.current_disk.nexus_index if self.db.current_disk else None
        for path, size, lastmod, hash, nexus in self.db.filter_files(
                pat, sql='SELECT ft.virtual_path, ft.size, ft.last_modified, ft.hash, ob.nexus FROM file_tree ft LEFT JOIN object ob ON ft.hash = ob.hash'):
            if destid is not None:
                if nexus is None:
                    continue
                if not disk_in_nexus(nexus, destid):
                    continue

            disks = ','.join(self.db.nexus_disks(nexus)) if nexus else ''
            print('%-7s (%s) %s' % (prettysize(size), disks, path))

    def get_tree(self):
        root = FileNode('', None, inode=1)
        nodebypath = {}

        max_copies = self.db.query_scalar('SELECT max(copies) + 1 FROM object')
        root.copy_files = [0] * max_copies
        root.copy_size = [0] * max_copies
        cnt = self.db.query_scalar('SELECT count(virtual_path) FROM file_tree')
        i = 0
        lprogress = ''

        vinode = 1
        hash_to_inode = {}

        sql = '''
        SELECT ft.last_modified, ft.hash, ft.size, ft.virtual_path, ft.priority, ft.maxcopies,
               ob.blocksize, ob.refs, ob.copies, ob.nexus, ob.last_path, ob.priority, ob.maxcopies
        FROM file_tree ft
        INNER JOIN object ob ON ft.hash = ob.hash
        '''

        progress = Progress.get_global()
        progress.set_status('loading db...')
        with self.db.execute(sql) as curs:
            for lastmod, hash, size, path, prio, maxcopies, blocksize, refs, copies, nexus, lpath, obprio, obmaxcopies in curs:
                cnode = root
                p = [root]
                for comp in path.split('/')[1:]:
                    nnode = cnode.children.get(comp)
                    if nnode is None:
                        vinode += 1
                        cnode.children[comp] = nnode = FileNode(comp, cnode, inode=vinode)
                        nnode.copy_size = [0] * max_copies
                        nnode.copy_files = [0] * max_copies
                    cnode = nnode
                    p.append(cnode)

                cnode.size = size
                cnode.lastmod = lastmod
                cnode.hash = hash
                cnode.priority = prio
                cnode.obj_priority = obprio
                cnode.maxcopies = maxcopies
                cnode.obj_maxcopies = obmaxcopies

                cnode.iscanon = path == lpath
                cnode.refs = refs
                cnode.copies = copies
                cnode.nexus = nexus
                inode = hash_to_inode.get(hash)
                if inode is None:
                    hash_to_inode[hash] = cnode.inode
                else:
                    cnode.inode = inode

                ref_blocksize = blocksize // refs
                if path == lpath:
                    ref_blocksize += blocksize % refs

                for node in p:
                    node.total_size += ref_blocksize
                    node.total_files += 1
                    node.copy_size[copies] += ref_blocksize
                    node.copy_files[copies] += 1

                i += 1
                progress.progress((i + 1) * 100 / cnt)

        progress.reset()
        return root

    def make_ncdu(self):
        root = self.get_tree()
        def visit_node(node):
            name = node.name.strip('/') or '/distbackup'
            if node.nexus:
                name += ' (' + ','.join(self.db.nexus_disks(node.nexus)) + ')'
            if node.priority != 1:
                name += f' (P{node.priority})'

            if node.maxcopies:
                name += f' (M{node.maxcopies})'

            data = {
                "name": name,
                "asize": node.size,
                "dsize": round_blocks(node.size),
            }

            if node.children:
                data['asize'] = 4096
                data['dsize'] = 4096
                data['dev'] = 1
                data['ino'] = node.inode
                rval = [data]
                for cnode in sorted(list(node.children.values()), key=lambda c:-c.total_size):
                    rval.append(visit_node(cnode))
            else:
                if node.refs > 1:
                    data['hlnkc'] = True
                data['ino'] = node.inode
                rval = data

            return rval
        data = [1, 1,
            {
                "progname": "ncdu",
                "progver": "1.15.1",
                "timestamp": int(time.time())
            },
            visit_node(root)
        ]
        return data



class FileChanged(Exception):
    '''Raised when a file changes during an update'''

class IncompleteUpdateError(Exception):
    '''Raised when a backup is requested when there are files with unknown hashes.'''

class ArgumentError(Exception):
    '''Raised when an command line argument is invalid'''


class DiskNotFoundError(ArgumentError):
    '''Raised when a backup disk cannot be found'''



import traceback
import time
import sqlite3
import uuid
import json
import logging

from typing import Optional
from pathlib import Path

# Third party imports
from ktpanda.sqlite_helper import SQLiteDB, in_transaction

from .constants import DBNAME
from .disk import BackupDisk
from .nexus import Nexus
from .utils import addcomma, canonical_path, compact_json, progress, pattern_to_regex
from .errors import ArgumentError, DiskNotFoundError

log = logging.getLogger(__name__)

DEFAULT_CONFIG = {
    'lock': False,
    'hide': False,
    'follow_symlinks': False,
    'include': [],
    'exclude': [],
}

SCHEMA_VERSION = 26

TABLES_INDICES = [
    # path_map: Maps virtual paths to actual directories on fhe filesystem
    """
    CREATE TABLE IF NOT EXISTS path_map (
        virtual_path TEXT PRIMARY KEY,
        real_path TEXT NOT NULL
    ) WITHOUT ROWID
    """,

    # path_config: Contains configuration options for individual directories
    """
    CREATE TABLE IF NOT EXISTS path_config (
        virtual_path TEXT PRIMARY KEY,
        config TEXT NOT NULL DEFAULT '{}'
    ) WITHOUT ROWID
    """,

    # file_tree: Contains information for each file that is backed up. Only leaf nodes
    # (i.e. files, not directories) have an entry in this table.
    """
    CREATE TABLE IF NOT EXISTS file_tree (
        -- Virtual path of the file, separated by "/"
        virtual_path TEXT PRIMARY KEY,

        -- Unix time in milliseconds
        last_modified INTEGER,

        -- SHA256 hash of the data. Empty if the hash is unknown, NULL if file is empty
        hash TEXT,

        -- Inode of the file (stat.st_ino). Used to quickly detect rename / move without having to rehash
        ino INTEGER,

        -- Size of the file, in bytes
        size INTEGER,

        -- Priority of the file. Files with a greater value for `priority` will be copied
        -- before lower priority files (assuming levels are equal)
        priority INTEGER NOT NULL DEFAULT 1,

        -- The maximum number of copies that this file can have on backup disks.
        maxcopies INTEGER,

        -- Compacted JSON of `.meta-inf` file associated with this file.
        metadata TEXT NOT NULL DEFAULT ''
    ) WITHOUT ROWID
    """,

    "CREATE INDEX IF NOT EXISTS file_tree_hash ON file_tree(hash)",
    "CREATE INDEX IF NOT EXISTS file_need_hash ON file_tree(virtual_path) WHERE hash = ''",

    # symbolic_link: Contains all symbolic links from the backup set.
    """
    CREATE TABLE IF NOT EXISTS symbolic_link (
        virtual_path TEXT PRIMARY KEY,
        target TEXT NOT NULL
    ) WITHOUT ROWID
    """,

    # nexus: A subset of the backup disks. Each row tracks the number and size of objects
    # stored on exactly those disks. Updated by triggers. The key is simply a binary
    # string of disks in little-endian order, stripped of trailing zeroes. '' represents
    # the empty set; '1' represents the nexus of only disk 0; '101' represents the nexus
    # of disk 0 with disk 2, etc.
    """
    CREATE TABLE IF NOT EXISTS nexus (
        -- Nexus key string, identifying which disks are in the set
        nexus TEXT PRIMARY KEY,

        -- Number of disks in the set
        level INTEGER GENERATED ALWAYS AS (length(replace(nexus, '0', ''))) STORED,

        -- Number of objects in this nexus. Updated by triggers.
        refs INTEGER DEFAULT 0,

        -- Sum of all `blocksize` columns of objects in this nexus. Updated by triggers.
        totalsize INTEGER DEFAULT 0,

        -- Names of disks in this nexus (informational, updated automatically)
        disks TEXT NOT NULL DEFAULT ''
    ) WITHOUT ROWID
    """,

    "CREATE INDEX IF NOT EXISTS nexus_prio ON nexus(level, totalsize DESC)",

    # object: Tracks which disks an object has been copied to. Multiple files with the
    # same content will reference the same object.
    """
    CREATE TABLE IF NOT EXISTS object (
        -- SHA256 hash of the object
        hash TEXT PRIMARY KEY,

        -- Exact size of the object in bytes
        size INTEGER DEFAULT 0,

        -- Size of the object, rounded up to next 4KB, plus 4KB for metadata
        blocksize INTEGER GENERATED ALWAYS AS (4096 + (size + 4095) & ~4095),

        -- Order of magnitude of the size
        size_order INTEGER GENERATED ALWAYS AS (CAST(floor(log2(blocksize + 0.01)) AS INTEGER)) STORED,

        -- Nexus of disks that this object is present on
        nexus TEXT NOT NULL DEFAULT '',

        -- Number of disks that this object exists on
        copies INTEGER GENERATED ALWAYS AS (length(replace(nexus, '0', ''))) STORED,

        -- Priority for this object. This is the maximum priority for any file_tree row
        -- that references this row.
        priority INTEGER NOT NULL DEFAULT 1,

        -- Maximum copies for this object. This is the maximum value for any file_tree row
        -- that references this row.
        maxcopies INTEGER,

        -- References to this object from the file_tree table, updated by triggers
        refs INTEGER NOT NULL DEFAULT 0,

        -- One of the paths that linked to this object. This is kept for
        -- informational purposes, so that if a file's content changes and causes
        -- this object to get deleted, we can show the user which file it was, along
        -- with the modification date.
        last_path TEXT,
        last_modtime INTEGER
    )
    """,

    "CREATE INDEX IF NOT EXISTS object_prio ON object(copies DESC, blocksize)",
    "CREATE INDEX IF NOT EXISTS object_size ON object(blocksize)",
    "CREATE INDEX IF NOT EXISTS object_refs ON object(refs)",
    "CREATE INDEX IF NOT EXISTS object_refs_copies ON object(refs, copies)",
    "CREATE INDEX IF NOT EXISTS object_nexus ON object(nexus)",

    # disk: Represents a backup disk that stores objects.
    """
    CREATE TABLE IF NOT EXISTS disk (
        -- UUID identifying the disk. This is generated when the disk is added and
        -- never changes.
        uuid TEXT PRIMARY KEY,

        -- Name of this disk. NULL if the disk has been deleted. Deleted disks are
        -- kept so the `nexus_index` can be reused.
        name TEXT UNIQUE,

        -- Index of this disk within nexus indetifiers.
        nexus_index INTEGER NOT NULL UNIQUE,

        -- Path to object files on the disk realtive to its mountpoint.
        relative_path TEXT,

        -- If not NULL, only search on partitions matching the given fstype
        fstype TEXT,

        -- If not NULL, only search on partitions matching the given UUID
        fsuuid TEXT,

        -- Size of the disk in bytes.
        size INTEGER

    ) WITHOUT ROWID
    """,

    "CREATE INDEX IF NOT EXISTS unused_nexus_index ON disk(nexus_index) WHERE name ISNULL",
]

TRIGGERS = [
    # Nexus update triggers. min(NEW.refs, 1) is used so that objects that aren't
    # referred to in the file_tree don't count as being present.
    """
    CREATE TRIGGER IF NOT EXISTS object_insert
    AFTER INSERT ON object
    BEGIN
        INSERT INTO nexus(nexus, refs, totalsize) VALUES(NEW.nexus, 1, (min(NEW.refs, 1) * NEW.blocksize))
        ON CONFLICT(nexus) DO UPDATE SET refs = refs + 1, totalsize = totalsize + (min(NEW.refs, 1) * NEW.blocksize);
    END
    """,

    """
    CREATE TRIGGER IF NOT EXISTS object_update
    AFTER UPDATE ON object
    WHEN OLD.blocksize != NEW.blocksize OR OLD.nexus != NEW.nexus OR OLD.refs != NEW.refs
    BEGIN
        INSERT INTO nexus(nexus, refs, totalsize) VALUES(NEW.nexus, 1, (min(NEW.refs, 1) * NEW.blocksize))
        ON CONFLICT(nexus) DO UPDATE SET refs = refs + 1, totalsize = totalsize + (min(NEW.refs, 1) * NEW.blocksize);

        UPDATE nexus SET refs = refs - 1, totalsize = totalsize - (min(OLD.refs, 1) * OLD.blocksize) WHERE nexus = OLD.nexus;
    END
    """,
    """
    CREATE TRIGGER IF NOT EXISTS object_delete
    BEFORE DELETE ON object
    BEGIN
        UPDATE nexus SET refs = refs - 1, totalsize = totalsize - (min(OLD.refs, 1) * OLD.blocksize) WHERE nexus = OLD.nexus;
    END
    """,

    """
    CREATE TRIGGER IF NOT EXISTS file_tree_insert
    AFTER INSERT ON file_tree
    WHEN NEW.hash <> ''
    BEGIN
        INSERT INTO object(hash, size, priority, maxcopies, refs, last_path, last_modtime)
          VALUES(NEW.hash, NEW.size, NEW.priority, NEW.maxcopies, NEW.virtual_path, 1, NEW.last_modified)
        ON CONFLICT(hash) DO UPDATE SET
          refs = refs + 1,
          priority = COALESCE((SELECT MAX(priority) FROM file_tree WHERE hash = NEW.hash), 1),
          maxcopies = (SELECT MAX(maxcopies) FROM file_tree WHERE hash = NEW.hash);
    END
    """,

    """
    CREATE TRIGGER IF NOT EXISTS file_tree_update
    AFTER UPDATE ON file_tree
    WHEN OLD.hash != NEW.hash
    BEGIN
        INSERT INTO object(hash, size, priority, maxcopies, refs, last_path, last_modtime)
          VALUES(NEW.hash, NEW.size, NEW.priority, NEW.maxcopies, NEW.virtual_path, 1, NEW.last_modified)
        ON CONFLICT(hash) DO UPDATE SET
          refs = refs + 1,
          priority = COALESCE((SELECT MAX(priority) FROM file_tree WHERE hash = NEW.hash), 1),
          maxcopies = (SELECT MAX(maxcopies) FROM file_tree WHERE hash = NEW.hash);

        UPDATE object SET
          refs = refs - 1,
          priority = COALESCE((SELECT MAX(priority) FROM file_tree WHERE hash = OLD.hash), 1),
          maxcopies = (SELECT MAX(maxcopies) FROM file_tree WHERE hash = OLD.hash)
        WHERE hash = OLD.hash;
    END
    """,

    """
    CREATE TRIGGER IF NOT EXISTS file_tree_delete
    BEFORE DELETE ON file_tree
    BEGIN
        UPDATE object SET
          refs = refs - 1,
          priority = COALESCE((SELECT MAX(priority) FROM file_tree WHERE hash = OLD.hash), 1),
          maxcopies = (SELECT MAX(maxcopies) FROM file_tree WHERE hash = OLD.hash)
        WHERE hash = OLD.hash;
    END
    """
]

PRE_DATA = [
    # Make sure there is always a nexus entry representing the empty set
    "INSERT OR IGNORE INTO nexus(nexus) VALUES ('')",
]

sql_funcs = {}
def sqlf(f):
    def ret(*args):
        try:
            return f(*args)
        except Exception:
            traceback.print_exc()
            raise
    sql_funcs[f.__name__] = ret
    return f

@sqlf
def disk_in_nexus(nexus, nexus_index):
    if nexus_index < 0 or nexus_index >= len(nexus):
        return False
    return nexus[nexus_index] == '1'

@sqlf
def nexus_with_disk(nexus, nexus_index):
    if nexus_index >= len(nexus):
        return nexus + '0' * (nexus_index - len(nexus)) + '1'
    return nexus[:nexus_index] + '1' + nexus[nexus_index + 1:]

@sqlf
def nexus_without_disk(nexus, nexus_index):
    if nexus_index >= len(nexus):
        return nexus
    return (nexus[:nexus_index] + '0' + nexus[nexus_index + 1:]).rstrip('0')

@sqlf
def nexus_level(nexus):
    return nexus.count('1')

@sqlf
def new_uuid():
    return str(uuid.uuid4())

@sqlf
def canonical_uuid(val):
    return str(uuid.UUID(val))

def prefix_sql(column, prefix):
    return f'({column} >= ? AND {column} < ?)', (prefix, prefix[:-1] + chr(1 + ord(prefix[-1:])))

class BackupDB(SQLiteDB):
    common_schema = TABLES_INDICES + TRIGGERS + PRE_DATA
    schema_version = SCHEMA_VERSION

    def __init__(self, dbpath:Path, readonly:bool=False, verbose:bool=False):
        if dbpath.is_dir():
            dbpath = dbpath / DBNAME

        super().__init__(dbpath, readonly=readonly)

        self.verbose = verbose

        self.disk_by_index:dict[int, BackupDisk] = {}
        self.disk_by_uuid:dict[str, BackupDisk] = {}
        self.disk_by_name:dict[str, BackupDisk] = {}
        self._disk_name_index:list[tuple[str, int]] = []

        self.current_disk:Optional[BackupDisk] = None
        self.defer_queue:list[tuple] = []
        self.last_run_deferred:float = 0

        self.path_map:dict[str, Path] = {}
        self.path_cache:dict[str, Path] = {}
        #self.diropt_cache = {}

        self.nocopydb:bool = False

    ##########################################################################################
    # Upgrades
    ##########################################################################################

    def _upgrade_to_13(self, oldvers):
        self.backend.execute('ALTER TABLE disks ADD COLUMN relative_path TEXT')

    def _upgrade_to_16(self, oldvers):
        self.backend.execute("ALTER TABLE file_names ADD COLUMN metadata TEXT NOT NULL DEFAULT ''")

    def _upgrade_to_18(self, oldvers):
        for column in ('file_names.path', 'path_map.virtual_path', 'path_config.virtual_path', 'links.path', 'files.last_path'):
            table, column = column.split('.')
            sql = f"UPDATE {table} SET {column} = '/' || {column} WHERE {column} NOT LIKE '/%'"
            self.backend.execute(sql)

    def _upgrade_to_20(self, oldvers):
        if self.backend.execute("SELECT name FROM sqlite_master WHERE name = 'source_path'"):
            self.backend.execute('ALTER TABLE source_path RENAME TO path_map')

        if self.backend.execute("SELECT name FROM sqlite_master WHERE name = 'source_config'"):
            self.backend.execute('ALTER TABLE source_config RENAME TO path_config')

    def _drop_triggers(self):
        triggers = [row[0] for row in self.backend.execute("SELECT name FROM sqlite_master WHERE type = 'trigger'")]
        for trigger in triggers:
            self.backend.execute(f'DROP TRIGGER {trigger}')

    def _upgrade_to_21(self, oldvers):
        '''*BIG UPDATE* - drop all triggers, create the new tables, copy
        the data over, then let it create the new triggers'''

        self._drop_triggers()

        self.exec_schema(TABLES_INDICES)

        self.backend.create_function('newid', 0, new_uuid)
        self.backend.execute('INSERT INTO object(hash, size, nexus, refs, last_path, last_modtime) SELECT hash, size, nexus, refs, last_path, last_lastmod FROM files')
        self.backend.execute('INSERT INTO file_tree(virtual_path, last_modified, hash, ino, size, metadata) SELECT path, lastmod, hash, ino, size, metadata FROM file_names')
        self.backend.execute('INSERT INTO symbolic_link(virtual_path, target) SELECT path, dest FROM links')
        self.backend.execute('INSERT INTO disk(uuid, name, nexus_index, relative_path, size) SELECT IFNULL(uuid, newid()), name, id - 1, relative_path, size FROM disks')
        self.backend.execute('INSERT INTO nexus(nexus, refs, totalsize) SELECT nexus, refs, totalsize FROM dist_nexus')
        self.backend.execute('DROP TABLE files')
        self.backend.execute('DROP TABLE file_names')
        self.backend.execute('DROP TABLE links')
        self.backend.execute('DROP TABLE disks')
        self.backend.execute('DROP TABLE dist_nexus')
        self.backend.execute('DROP TABLE IF EXISTS vars')
        self.backend.execute('DROP TABLE IF EXISTS source_lock')

    def _upgrade_to_22(self, oldvers):
        self._drop_triggers()
        self.backend.execute("ALTER TABLE file_tree ADD COLUMN maxcopies INTEGER")
        self.backend.execute("ALTER TABLE object ADD COLUMN maxcopies INTEGER")

    def _upgrade_to_23(self, oldvers):
        self._drop_triggers()
        self.backend.execute("ALTER TABLE nexus RENAME COLUMN count TO level")

    def _upgrade_to_24(self, oldvers):
        self.backend.execute("ALTER TABLE nexus ADD COLUMN disks TEXT NOT NULL DEFAULT ''")
        self.update_nexus_disk_text()

    def _upgrade_to_25(self, oldvers):
        self._drop_triggers()
        self.backend.execute('ALTER TABLE object RENAME TO old_object')
        self.backend.execute('ALTER TABLE file_tree RENAME TO old_file_tree')
        self.exec_schema(TABLES_INDICES)
        self.backend.execute('INSERT INTO object(hash, size, nexus, refs, priority, last_path, last_modtime) SELECT hash, size, nexus, refs, priority, last_path, last_modtime FROM old_object')
        self.backend.execute('INSERT INTO file_tree(virtual_path, last_modified, hash, ino, size, priority, metadata) SELECT virtual_path, last_modified, hash, ino, size, priority, metadata FROM old_file_tree')
        self.backend.execute('DROP TABLE old_object')
        self.backend.execute('DROP TABLE old_file_tree')

    def _upgrade_to_26(self, oldvers):
        self._drop_triggers()

    ##########################################################################################
    # Database
    ##########################################################################################

    def connect(self, *a, **kw):
        super().connect(*a, **kw)
        for n, f in sql_funcs.items():
            self.backend.create_function(n, -1, f)

        self.disk_by_index = {}
        self.disk_by_name = {}
        self.disk_by_uuid = {}
        with self.execute('SELECT uuid, name, nexus_index, relative_path, size, fstype, fsuuid FROM disk WHERE name NOTNULL') as curs:
            for uuid, name, nexus_index, relative_path, size, fstype, fsuuid in curs:
                disk = BackupDisk(uuid, name, nexus_index, relative_path, size, fstype, fsuuid)
                self.disk_by_uuid[uuid] = disk
                self.disk_by_index[nexus_index] = disk
                self.disk_by_name[name] = disk

        self._disk_name_index = sorted((disk.name, index) for index, disk in self.disk_by_index.items())

        self.path_map = {
            virtual_path: Path(real_path)
            for virtual_path, real_path in
            self.query_list('SELECT virtual_path, real_path FROM path_map')
        }
        self.path_cache = self.path_map.copy()

    @in_transaction
    def copy_database(self, dest_dir=None):
        if dest_dir is None:
            dest_dir = self.current_disk.data_path

        self.execute_nonquery('PRAGMA wal_checkpoint(FULL)')
        if self.nocopydb:
            return

        if not self.current_disk or not self.current_disk.data_path:
            return

        log.info('copying database...')
        tmppath = dest_dir / (DBNAME + '.new')
        backup_db = sqlite3.connect(tmppath)
        def log_progress(status, remaining, total):
            progress((total - remaining) * 100 / total)

        self.backend.backup(backup_db, pages=256, progress=log_progress)
        backup_db.commit()
        backup_db.execute('PRAGMA wal_checkpoint(FULL)')
        backup_db.close()
        progress(None)

        dstf = tmppath.with_name(DBNAME)
        tmppath.replace(dstf)

    def defer(self, f, *a, **kw):
        '''Queue a database-modifying function to run in the near future in a transaction.'''
        self.defer_queue.append((f, a, kw))

    @in_transaction
    def _run_deferred(self):
        for f, a, kw in self.defer_queue:
            f(*a, **kw)
        del self.defer_queue[:]

    def check_run_deferred(self, force=False):
        ctime = time.time()
        if force or ctime >= self.last_run_deferred + 0.2:
            self.last_run_deferred = ctime
            if self.defer_queue:
                self._run_deferred()

    def load_nexus(self):
        rtn = {}
        with self.execute('SELECT nexus, level, refs, totalsize, disks FROM nexus') as curs:
            for nexus, level, refs, totalsize, disks in curs:
                rtn[nexus] = Nexus(nexus, level, refs, totalsize, disks)
        return rtn

    def create_empty_nexus(self, nexusid:str):
        return Nexus(nexusid, nexus_level(nexusid), 0, 0, ','.join(self.nexus_disks(nexusid)))

    def get_used_space(self, nexus_index):
        return self.query_scalar('SELECT COALESCE(sum(totalsize), 0) FROM nexus WHERE disk_in_nexus(nexus, ?)', (nexus_index,))

    def update_free_space(self):
        self.current_disk.used_space = self.get_used_space(self.current_disk.nexus_index)

    ##########################################################################################
    # Disk selection
    ##########################################################################################

    def set_backup_disk(self, arg):
        if '/' in arg:
            path = Path(arg)
            if not path.is_dir():
                raise DiskNotFoundError(f'path {path} does not exist or is not a directory')

            rdisk = BackupDisk.read_config(path)
            if not rdisk:
                raise DiskNotFoundError(f'path {path} does not contain a valid distbackup config')

            try:
                disk = self.disk_by_uuid[rdisk.uuid]
            except KeyError:
                raise ArgumentError(f'path {path} is from a different backup set') from None

            disk.data_path = path
            if disk.name != rdisk.name:
                disk.write_config()

            return self._set_current_disk(disk)

        try:
            return self.set_backup_disk_from_uuid(arg)
        except (ValueError, DiskNotFoundError):
            pass

        return self.set_backup_disk_from_name(arg)

    def set_backup_disk_from_name(self, name, search_data=True):
        try:
            disk = self.disk_by_name[name]
        except KeyError:
            raise DiskNotFoundError(f'disk {name} not found') from None

        self._set_current_disk(disk, search_data)

    def set_backup_disk_from_uuid(self, diskuuid, search_data=True):
        try:
            disk = self.disk_by_uuid[str(uuid.UUID(diskuuid))]
        except KeyError:
            raise DiskNotFoundError(f'disk {diskuuid} not found') from None

        self._set_current_disk(disk, search_data)

    def _set_current_disk(self, disk, search_data=True):
        self.current_disk = disk
        if not disk.data_path and search_data:
            BackupDisk.find_disk_data_paths([disk])

        log.debug(f'using data path {disk.data_path}')
        self.update_free_space()


    def require_disk(self):
        if not self.current_disk:
            raise ArgumentError('no disk selected')

    def require_disk_path(self):
        self.require_disk()
        if not self.current_disk.data_path:
            raise ArgumentError(f'{self.current_disk.name} is not mounted, or not configured correctly')

    def find_disk_data_paths(self):
        BackupDisk.find_disk_data_paths(self.disk_by_uuid.values())

    ##########################################################################################
    # Disk management
    ##########################################################################################

    @in_transaction('IMMEDIATE')
    def add_disk(self, name, path, size=None):
        uuid = new_uuid()
        new_disk = BackupDisk(uuid, name, None, size)
        new_disk.set_from_path(path)

        try:
            nexus_index = self.query_scalar('SELECT COALESCE((SELECT MIN(nexus_index) FROM disk WHERE name ISNULL), (SELECT MAX(nexus_index) + 1 FROM disk), 0)')
            new_disk.nexus_index = nexus_index
            self.execute_nonquery('DELETE FROM disk WHERE nexus_index = ?', (nexus_index,))
            self.execute_nonquery(
                'INSERT INTO disk(uuid, name, nexus_index, size, relative_path, fstype, fsuuid) VALUES(?, ?, ?, ?, ?, ?, ?)',
                (uuid, name, nexus_index, new_disk.size, new_disk.relative_path, new_disk.fstype, new_disk.fsuuid))

        except sqlite3.IntegrityError as e:
            raise ArgumentError(f'disk {name} already exists ({e})') from None

        self.current_disk = new_disk

        self.disk_by_index[nexus_index] = new_disk
        self.disk_by_name[name] = new_disk
        self.disk_by_uuid[uuid] = new_disk
        self._disk_name_index = sorted((disk.name, index) for index, disk in self.disk_by_index.items())

        self.update_nexus_disk_text()
        self.update_free_space()
        new_disk.write_config()

        return new_disk


    @in_transaction
    def rename_disk(self, newname):
        self.require_disk()
        self.execute_nonquery('UPDATE disk SET name = ? WHERE uuid = ?', (newname, self.current_disk.uuid))
        self.disk_by_name.pop(self.current_disk.name, None)
        self.current_disk.name = newname
        self.disk_by_name[newname] = self.current_disk

        self.update_nexus_disk_text()

        if self.current_disk.data_path:
            self.current_disk.write_config()

    @in_transaction
    def set_disk_params(self, size=None, relpath=None, path=None, fstype=None, fsuuid=None):
        disk = self.current_disk
        if path is not None:
            disk.set_from_path(path)
            self.execute_nonquery(
                'UPDATE disk SET relative_path= ?, fstype = ?, fsuuid = ? WHERE uuid = ?',
                (disk.relative_path, disk.fstype, disk.fsuuid, disk.uuid)
            )

        if size is not None:
            self.execute_nonquery('UPDATE disk SET size = ? WHERE uuid = ?', (size, disk.uuid))
            disk.size = size

        if relpath is not None:
            disk.relative_path = relpath or None
            self.execute_nonquery('UPDATE disk SET relative_path = ? WHERE uuid = ?', (disk.relative_path, disk.uuid))

        if fstype is not None:
            disk.fstype = fstype or None
            self.execute_nonquery('UPDATE disk SET fstype = ? WHERE uuid = ?', (disk.fstype, disk.uuid))

        if fsuuid is not None:
            disk.fsuuid = fsuuid or None
            self.execute_nonquery('UPDATE disk SET fsuuid = ? WHERE uuid = ?', (disk.fsuuid, disk.uuid))

        self.update_nexus_disk_text()

        if disk.data_path:
            disk.write_config()

    def drop_disk_objects(self):
        index = self.current_disk.nexus_index
        with self.execute('SELECT nexus, level, totalsize FROM nexus') as curs:
            for nexus, copies, totalsize in curs:
                if disk_in_nexus(nexus, index):
                    new_nexus = nexus_without_disk(nexus, index)
                    self.execute_nonquery('UPDATE object SET nexus = ? WHERE nexus = ?', (new_nexus, nexus))

    @in_transaction
    def drop_disk(self):
        self.require_disk()
        self.drop_disk_objects()
        self.cleanrefs()

    @in_transaction
    def delete_disk(self):
        self.require_disk()

        used_space = self.query_scalar('SELECT sum(totalsize) FROM nexus WHERE disk_in_nexus(nexus, ?)', (self.current_disk.nexus_index,))
        if used_space:
            print(f'{self.current_disk.name} has backup files (use `disk drop`)')
            return

        self.execute_nonquery('UPDATE disk SET name = NULL WHERE uuid = ?', (self.current_disk.uuid,))
        self.cleanrefs()

    @in_transaction
    def _refresh_disk(self):
        self.require_disk()
        self.current_disk.migrate_data()
        self.drop_disk_objects()
        for hash, fpath, stat in self.current_disk.walk_data():
            log.info(f'found {hash}')
            self.execute_nonquery('INSERT OR IGNORE INTO object(hash, size) VALUES(?, ?)', (hash, stat.st_size))
            self.set_object_on_disk(hash, self.current_disk.nexus_index, True)

        self.cleanrefs()

    def refresh_disk(self):
        self._refresh_disk()
        self.update_free_space()
        self.copy_database()

    ##########################################################################################
    # Configuration
    ##########################################################################################

    @in_transaction
    def map_source(self, vpath:str, ppath:str):
        self.execute_nonquery('INSERT OR REPLACE INTO path_map(virtual_path, real_path) VALUES(?, ?)', (vpath, ppath))

    @in_transaction
    def unmap_source(self, vpath:str):
        rowcount, lastrowid = self.execute_nonquery('DELETE FROM path_map WHERE virtual_path = ?', (vpath,))
        return rowcount > 0

    @in_transaction
    def set_path_config(self, paths:list[str], updates:dict, clear:bool=False, append_keys=()):
        rtn = []
        for vpath in paths:
            vpath = canonical_path(vpath, True)
            prev_config_text = self.query_scalar('SELECT config FROM path_config WHERE virtual_path = ?', (vpath,))
            prev_config = json.loads(prev_config_text) if prev_config_text else {}

            if prev_config_text is None:
                config = updates.copy()
                if config:
                    self.execute_nonquery('INSERT INTO path_config(virtual_path, config) VALUES(?, ?)', (vpath, compact_json(config)))
            else:
                config = {} if clear else prev_config.copy()
                for key, val in updates.items():
                    if val is None or val == DEFAULT_CONFIG.get(key):
                        config.pop(key, None)
                    else:
                        if key in append_keys:
                            newval = list(config.get(key) or ())
                            for cval in val:
                                if cval not in newval:
                                    newval.append(cval)
                            val = newval
                        config[key] = val
                if config:
                    self.execute_nonquery('UPDATE path_config SET config = ? WHERE virtual_path = ?', (compact_json(config), vpath))
                else:
                    self.execute_nonquery("DELETE FROM path_config WHERE virtual_path = ?", (vpath,))

            rtn.append((vpath, prev_config, config))
        return rtn

    def export_config(self, path_prefix=None):
        path_config = {}
        disks = {}
        data = {
            'path_map': {
                k: str(v) for k, v in self.path_map.items()
            }
        }

        sql = 'SELECT virtual_path, config FROM path_config'
        args = ()
        if path_prefix:
            # Make sure the specified path is in the resulting export
            path_config[path_prefix] = {}
            text, args = prefix_sql('virtual_path', path_prefix)
            sql += " WHERE " + text

            data['path_prefix'] = path_prefix

        data['path_config'] = path_config
        for path, config in self.execute(sql, args):
            path_config[path] = json.loads(config)

        data['disks'] = disks
        with self.execute('SELECT uuid, name, relative_path, size, fstype, fsuuid FROM disk WHERE name NOTNULL') as curs:
            for uuid, name, relative_path, size, fstype, fsuuid in curs:
                disks[uuid] = {
                    'name': name,
                    'size_bytes': addcomma(size),
                    'relative_path': relative_path,
                    'fstype': fstype,
                    'fsuuid': fsuuid
                }

        return data

    @in_transaction
    def import_config(self, data):
        if 'path_map' in data:
            self.execute_nonquery('DELETE FROM path_map')
            for virtual_path, real_path in data['path_map'].items():
                if not isinstance(real_path, str):
                    raise ValueError(f'Expected string for path map: {virtual_path!r}, got {type(real_path).__name__!r}')

                virtual_path = canonical_path(virtual_path, True)
                self.execute_nonquery('INSERT INTO path_map(virtual_path, real_path) VALUES(?, ?)', (virtual_path, real_path))

        if 'path_config' in data:
            path_prefix = data.get('path_prefix')
            if path_prefix:
                text, args = prefix_sql('virtual_path', path_prefix)

                self.execute_nonquery("DELETE FROM path_config WHERE " + text, args)
            else:
                self.execute_nonquery('DELETE FROM path_config')

            for virtual_path, config in data['path_config'].items():
                virtual_path = canonical_path(virtual_path, True)
                if not isinstance(config, dict):
                    raise ValueError(f'Expected dict for path config: {virtual_path!r}, got {type(config).__name__}')

                for key, val in DEFAULT_CONFIG.items():
                    if key in config and type(val) != type(config[key]):
                        raise ValueError(f'Path {virtual_path}: Expected {type(val).__name__} for {key}, got {type(config[key]).__name__}')

                self.execute_nonquery('INSERT INTO path_config(virtual_path, config) VALUES(?, ?)', (virtual_path, compact_json(config)))

        if 'disks' in data:
            for uuid, config in data['disks'].items():
                realuuid = self.query_scalar('SELECT uuid FROM disk WHERE uuid = ?', (canonical_uuid(uuid),))
                if realuuid is None:
                    raise ValueError(f'Disk with UUID {realuuid} not in database (disks cannot be added with import)')

                if 'size_bytes' in config:
                    size = int(str(config['size_bytes']).replace(',', ''))
                    self.execute_nonquery('UPDATE disk SET size = ? WHERE uuid = ?', (size, realuuid))

                if 'relative_path' in config:
                    val = config['relative_path']
                    self.execute_nonquery('UPDATE disk SET relative_path = ? WHERE uuid = ?', (val, realuuid))

                if 'fstype' in config:
                    val = config['fstype']
                    self.execute_nonquery('UPDATE disk SET fstype = ? WHERE uuid = ?', (val, realuuid))

                if 'fsuuid' in config:
                    val = config['fsuuid']
                    self.execute_nonquery('UPDATE disk SET fsuuid = ? WHERE uuid = ?', (val, realuuid))

                if 'name' in config:
                    val = config['name']
                    self.execute_nonquery('UPDATE disk SET name = ? WHERE uuid = ?', (val, realuuid))

    ##########################################################################################
    # Utils
    ##########################################################################################

    def nexus_disks(self, nexus):
        '''Given a nexus ID, yields a sequence of the names of the disks in the nexus.'''
        for name, index in self._disk_name_index:
            if disk_in_nexus(nexus, index):
                yield name

    @in_transaction('IMMEDIATE')
    def cleanrefs(self):
        if not self.readonly:
            self.execute_nonquery('DELETE FROM nexus WHERE refs = 0 AND nexus != \'\'')
            self.execute_nonquery('DELETE FROM object WHERE refs = 0 AND copies = 0')

    @in_transaction
    def update_nexus_disk_text(self):
        updates = []
        disk_names = {}
        with self.execute('SELECT name, nexus_index FROM disk WHERE name NOTNULL') as curs:
            for name, index in curs:
                disk_names[index] = name

        with self.execute('SELECT nexus, disks FROM nexus') as curs:
            for nexus, text in curs:
                new_text = ','.join(sorted(name for i, c in enumerate(nexus) if (c == '1' and (name := disk_names.get(i)))))
                if new_text != text:
                    updates.append((new_text, nexus))
        if updates:
            with self.executemany('UPDATE nexus SET disks = ? WHERE nexus = ?', updates) as curs:
                pass

    def filter_files(self, pat=None, existing=(), sql='SELECT ft.virtual_path, ft.last_modified, ft.hash, ob.nexus FROM file_tree ft LEFT JOIN object ob ON ft.hash = ob.hash'):
        rx, prefix = pattern_to_regex(pat)

        args = ()
        if prefix:
            text, args = prefix_sql('ft.virtual_path', prefix)
            sql += r" WHERE " + text

        with self.execute(sql, args) as curs:
            for row in curs:
                path = row[0]
                if rx and not rx.match(path):
                    continue

                if any((base / path).exists() for base in existing):
                    continue

                yield row

    @in_transaction
    def set_object_on_disk(self, hash, diskid, present):
        prev_nexus = self.query_scalar('SELECT nexus FROM object WHERE hash = ?', (hash,))
        nexus = nexus_with_disk(prev_nexus, diskid) if present else nexus_without_disk(prev_nexus, diskid)
        if nexus != prev_nexus:
            self.execute_nonquery('UPDATE object SET nexus = ? WHERE hash = ?', (nexus, hash))
        return nexus

    @in_transaction
    def set_file_hash(self, vpath, hash):
        self.execute_nonquery('UPDATE file_tree SET hash = ? WHERE virtual_path = ?', (hash, vpath))

    ##########################################################################################
    # Path map
    ##########################################################################################

    def iter_paths(self):
        return self.path_map.items()

    def get_mapped_path(self, vpath:str):
        return self.path_map.get(vpath)

    def get_dir_path(self, dir):
        path = self.path_cache.get(dir + '/')
        if path is None:
            if not dir:
                return None
            parent, _, dirname = dir.rpartition('/')
            parent_path = self.get_dir_path(parent)
            if parent_path is None:
                return None
            path = self.path_cache[dir] = parent_path / dirname
        return path

    def get_file_path(self, fn):
        dir, _, filename = fn.rpartition('/')
        dir_path = self.get_dir_path(dir)
        if dir_path is None:
            return None
        return dir_path / filename



import json
import fnmatch
import os
import re
import time
import logging

from typing import Optional
from collections import deque
from stat import S_ISREG, S_ISLNK, S_ISDIR

from ktpanda.sqlite_helper import in_transaction

from .constants import META_SUFFIX
from .errors import FileChanged
from .utils import int_mtime, compact_json, Progress
from .database import BackupDB
from .sourcefile import SourceFile

log = logging.getLogger(__name__)

class Matcher:
    RX_PATTERN = re.compile(r'^(!*)\s*([:\?/]?)(.*)')
    def __init__(self):
        self.patterns = []
        self.results = {}
        self.regex = None

    def add_pattern(self, pattern, result):
        key = f'p{len(self.patterns)}'
        self.patterns.append(f'{pattern}(?P<{key}>)')
        self.results[key] = result
        self.regex = None

    def parse_pattern(self, pattern):
        m = self.RX_PATTERN.match(pattern)
        prio, type, pat = m.groups()
        prio = len(prio)

        if type == ':':
            pat = r'\A' + re.escape(pat) + r'\Z'
        elif type == '?':
            pat = r'\A' + fnmatch.translate(pat)
        elif type  == '/':
            # Compile it now to validate it, even though we're just extracting the pattern
            # to combine later
            pat = re.compile(pat).pattern
        else:
            if re.search(r'[\*\?\[]', pat):
                pat = r'\A' + fnmatch.translate(pat)
            else:
                pat = r'\A' + re.escape(pat) + r'\Z'
        self.add_pattern(pat, prio)

    def match(self, fn):
        if self.regex is None:
            self.regex = re.compile('|'.join(self.patterns), re.S)
        m = self.regex.search(fn)
        if not m:
            return None
        return self.results[m.lastgroup]

    @classmethod
    def parse(cls, seq):
        self = cls()
        for pat in seq:
            self.parse_pattern(pat)
        return self

    @staticmethod
    def match_any(matchers, fn):
        for matcher in matchers:
            res = matcher.match(fn)
            if res is not None:
                return res
        return None

class Updater:
    def __init__(self, db:BackupDB, ignore_lock=False):
        self.db:BackupDB = db
        self.need_hash:list = []
        self.row_by_path:Optional[dict[str, tuple]] = None
        self.row_by_ino:Optional[dict[tuple, tuple]] = None
        self.ignore_lock:bool = ignore_lock
        self.symlink_unseen:set = set()

    def add_link(self, src, dest):
        if not self.db.readonly:
            self.db.execute_nonquery('INSERT OR REPLACE INTO symbolic_link(virtual_path, target) VALUES(?, ?)', (src, dest))
        self.symlink_unseen.pop(src, None)

    def update_file(self, virtual_path, srcf):
        oldhash = None
        fpath = srcf.path
        oldrow = self.row_by_path.pop(virtual_path, None)
        if not S_ISREG(srcf.mode):
            if oldrow:
                self.row_by_path[virtual_path] = oldrow

            if S_ISLNK(srcf.mode):
                dest = None
                try:
                    dest = os.readlink(fpath)
                except OSError:
                    pass
                if dest:
                    self.add_link(virtual_path, dest)

            return

        if srcf.priority == 0:
            return

        need_update_objects = False
        need_update_file_tree = False

        mtime = int_mtime(srcf.mtime)
        metatext = compact_json(srcf.metadata) if srcf.metadata else ''
        ometa = ''
        oprio = None

        ## Did the path already exist in the database?
        if oldrow:
            ## Is any of the data different?
            if oldrow[0] != mtime or oldrow[2] != srcf.size or oldrow[1] is None:
                need_update_file_tree = True

                ## Is the time different enough to need to rehash?
                if abs(oldrow[0] - mtime) >= 1000:
                    oldrow = None
        else:
            need_update_file_tree = True
            ## Check for a rename or hardlink.
            oldrow = self.row_by_ino.get((srcf.inode, srcf.size, mtime))

        if oldrow:
            omtime, oldhash, csize, opath, oino, ometa, oprio, omaxcopies = oldrow
            if opath != virtual_path:
                log.info(f'detect rename/hardlink {opath} -> {virtual_path}')

        if not oldhash:
            need_update_objects = True
            need_update_file_tree = True

            if srcf.hash is not None:
                log.info(f'hash {virtual_path} (precomputed)')
            elif srcf.size:
                log.info(f'need hash for {virtual_path} {srcf.path}')
                srcf.hash = ''
                self.need_hash.append((virtual_path, srcf))
        else:
            srcf.hash = oldhash

        if metatext != ometa or srcf.priority != oprio or srcf.maxcopies != omaxcopies:
            need_update_file_tree = True

        if srcf.size:
            if need_update_objects and srcf.hash and not self.db.readonly:
                self.db.execute_nonquery('INSERT OR IGNORE INTO object(hash) VALUES(?)', (srcf.hash,))
                self.db.execute_nonquery('UPDATE object SET size = ?, last_path = ?, last_modtime = ? WHERE hash = ?',
                             (srcf.size, virtual_path, mtime, srcf.hash))
        else:
            srcf.hash = None

        if need_update_file_tree and not self.db.readonly:
            self.db.execute_nonquery('INSERT OR REPLACE INTO file_tree(virtual_path, last_modified, hash, ino, size, priority, maxcopies, metadata) VALUES(?, ?, ?, ?, ?, ?, ?, ?)',
                         (virtual_path, int_mtime(srcf.mtime), srcf.hash, srcf.inode, srcf.size, srcf.priority or 1, srcf.maxcopies, metatext))

    def lock_path(self, vpath):
        log.info(f'{vpath} is locked')

        links = self.symlink_unseen
        locked_links = [k for k in links if k.startswith(vpath)]
        for k in locked_links:
            del links[k]

        bypath = self.row_by_path
        locked_files = [k for k in bypath if k.startswith(vpath)]
        for k in locked_files:
            del bypath[k]

    def walk_source(self):
        dq = deque()
        for virtual_path, real_path in self.db.iter_paths():
            dq.append((virtual_path, real_path, (), 1, None))

        while dq:
            vpath, apath, matchers, dir_priority, dir_maxcopies = dq.popleft()
            config = self.db.query_scalar('SELECT config FROM path_config WHERE virtual_path = ?', (vpath,))

            config = {} if config is None else json.loads(config)

            if not self.ignore_lock and config.get('lock'):
                self.lock_path(vpath)
                continue

            if config.get('hide'):
                continue

            config_prio = config.get('priority')
            if config_prio is not None:
                dir_priority = config_prio

            maxcopies = config.get('maxcopies')
            if maxcopies is not None:
                dir_maxcopies = maxcopies

            if dir_priority == 0:
                continue

            if (apath / ('.distbackup.skip')).exists():
                continue

            if config.get('noinherit', False):
                matchers = ()

            follow_links = config.get('follow_symlinks', False)

            rules = config.get('rules')
            if rules:
                matchers = (Matcher.parse(rules),) + matchers

            try:
                files = sorted(apath.iterdir())
            except EnvironmentError as e:
                log.warning(f'Cannot list {apath} ({e}), locking')
                self.lock_path(vpath)
                continue

            log.info(f'entering {vpath}')

            files.sort()
            for f in files:
                priority = Matcher.match_any(matchers, f.name)
                if priority == 0:
                    continue

                if priority is None:
                    priority = dir_priority

                virtual_path = f'{vpath.rstrip("/")}/{f.name}'
                if f.suffix == META_SUFFIX:
                    continue

                if self.db.get_mapped_path(virtual_path):
                    continue

                try:
                    srcf = SourceFile(f, f.stat() if follow_links else f.lstat())
                except os.error as e:
                    log.info(f'WARNING: cannot read {f} ({e})')
                    continue

                srcf.read_meta()
                if srcf.priority is None:
                    srcf.priority = priority

                if srcf.maxcopies is None:
                    srcf.maxcopies = dir_maxcopies

                while True:
                    self.update_file(virtual_path, srcf)
                    if S_ISDIR(srcf.mode):
                        dq.append((virtual_path + '/', f, matchers, priority, dir_maxcopies))
                    break

    def hash_files(self):
        self.db.check_run_deferred(True)

        try:
            for index, (virtual_path, srcf) in enumerate(self.need_hash):
                if self.db.readonly:
                    log.info(f'would hash {index + 1}/{len(self.need_hash)}: {virtual_path}')
                    continue

                while True:
                    log.info(f'hash {index + 1}/{len(self.need_hash)}: {virtual_path}')
                    try:
                        srcf.copy(self.db.check_run_deferred)
                        srcf.check_change()
                        break
                    except FileChanged:
                        log.info("file changed while reading!")
                        srcf.reset()
                        continue
                    except OSError as e:
                        log.info(f'WARNING: cannot read {virtual_path} ({e})')
                        break

                if srcf.hash:
                    self.db.defer(self.update_hash, virtual_path, srcf.hash, srcf.size, srcf.mtime)

                self.db.check_run_deferred()
        finally:
            self.db.check_run_deferred(True)

    def update_hash(self, virtual_path, hash, size, mtime):
        self.db.execute_nonquery('INSERT OR IGNORE INTO object(hash) VALUES(?)', (hash,))
        self.db.execute_nonquery('UPDATE object SET size = ?, last_path = ?, last_modtime = ? WHERE hash = ?',
                                 (size, virtual_path, mtime, hash))
        self.db.execute_nonquery('UPDATE file_tree SET hash = ? WHERE virtual_path = ?', (hash, virtual_path))

    def update_source(self, ignore_lock=False):
        self.need_hash = []
        self._update_source(ignore_lock)
        self.hash_files()

    @in_transaction(mode='IMMEDIATE', attr='db')
    def _update_source(self, ignore_lock=False):
        bypath = self.row_by_path = {}
        byino = self.row_by_ino = {}
        self.ignore_lock = ignore_lock

        cnt = self.db.query_scalar('SELECT count(virtual_path) FROM file_tree')
        i = 0
        t = time.time()
        progress = Progress.get_global()
        progress.set_status('loading db...')

        with self.db.execute('SELECT last_modified, hash, size, virtual_path, ino, metadata, priority, maxcopies FROM file_tree') as curs:
            for row in curs:
                bypath[row[3]] = row
                byino[(row[4], row[2], row[0])] = row

                i += 1
                progress.progress((i + 1) * 100 / cnt)

        with self.db.execute('SELECT virtual_path, target FROM symbolic_link') as curs:
            self.symlink_unseen = dict(curs)

        progress.reset()

        e = time.time()
        log.info(f'loaded in {e - t}s')

        self.walk_source()

        for path in sorted(self.row_by_path.keys()):
            log.info(f'deleted: {path}')

        if not self.db.readonly:
            self.db.executemany('DELETE FROM file_tree WHERE virtual_path = ?', ((k,) for k in self.row_by_path))
            self.db.executemany('DELETE FROM symbolic_link WHERE virtual_path = ?', ((k,) for k in self.symlink_unseen))
        self.db.cleanrefs()



import sys
import os
import json
import re
import uuid
import argparse
import fnmatch
import traceback
from ktpanda.textcolor import Colorizer

ROOT_ARG_PARSER = argparse.ArgumentParser(description='')

COMMANDS = {}

class CacheDict(dict):
    '''Works like `defaultdict`, except it passes the key to the factory function'''
    def __init__(self, factory):
        super().__init__()
        self.factory = factory

    def __missing__(self, key):
        r = self[key] = self.factory(key)
        return r

class Progress:
    _global_instance = None

    def __init__(self):
        self.current_percent:float = 0.0
        self.converted_progress:int = 0
        self.status:str = ''

    @classmethod
    def set_global(cls, progress):
        cls._global_instance = progress

    @classmethod
    def get_global(cls):
        return cls._global_instance or NullProgress()

    def set_status(self, status:str):
        self.status = status
        self.display()

    def _convert_percent(self, percent:float):
        return int(percent * 10)

    def progress(self, percent:float):
        if percent is None:
            self.reset()
            return

        new_progress = self._convert_percent(percent)
        if new_progress != self.converted_progress:
            self.current_percent = percent
            self.converted_progress = new_progress
            self.display()

    def reset(self):
        self.status = ''
        self.current_percent = 0.0
        self.converted_progress = self._convert_percent(0.0)
        self.clear()

    def display(self):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError

class NullProgress(Progress):
    def display(self):
        pass

    def clear(self):
        pass

class ProgressBar(Progress):
    def display(self):
        barsz = 90 - len(self.status) - 12
        barfill = int(self.current_percent * barsz / 100)
        text = f'\r\033[?7l {self.status} {self.current_percent:5.1f}% |{"=" * barfill}{" " * (barsz - barfill)}|\r\033[?7h'
        sys.stdout.write(text)
        sys.stdout.flush()

    def clear(self):
        sys.stdout.write('\r\033[K')
        sys.stdout.flush()

def progress(pct=None):
    if Progress._global_instance:
        Progress._global_instance.progress(pct)

class Command:
    def __init__(self, name, parserkw, func, parser):
        self.name = name
        self.func = func
        self.parserkw = parserkw

        self.parser = parser
        self.subparsers = None

    def get_subparsers(self):
        if self.subparsers is None:
            self.subparsers = self.parser.add_subparsers(
                title='Subcommands',
                help='Use %(prog)s <command> --help for subcommand usage',
                metavar='command'
            )
        return self.subparsers

COMMANDS[()] = Command((), {}, lambda db, parser, args: parser.print_help(), ROOT_ARG_PARSER)

def command(*name, **parserkw):
    def rtn(func):
        parent_cmd = COMMANDS[name[:-1]]
        if 'description' not in parserkw:
            parserkw['description'] = parserkw.get('help')
        parser = parent_cmd.get_subparsers().add_parser(name[-1], **parserkw)
        parser.set_defaults(command_name=name)
        for argf in reversed(getattr(func, 'argfuncs', ())):
            argf(parser)

        COMMANDS[name] = Command(name, parserkw, func, parser)
        return func
    return rtn

def _arg(f, argf):
    try:
        args = f.argfuncs
    except AttributeError:
        args = f.argfuncs = []
    args.append(argf)
    return f

def arg(*a, **kw):
    completer = kw.pop('completer', None)
    def argf(p):
        arg = p.add_argument(*a, **kw)
        if completer:
            arg.completer = completer
    return lambda f: _arg(f, argf)

_unit = list("BKMGTPE")
_unit_color = [
    'fff',
    '05f',
    '3f3',
    'fc0',
    'c0f',
    '0f9',
    '0f0',
]

def _prettysize(n):
    neg = n < 0
    if neg:
        n = -n
    un = 0
    while n >= 1000.0:
        n /= 1000.0
        un += 1
    pfx = ('-' if neg else '')
    if un:
        return pfx, format(n, '.2f'), un
    return pfx, format(n, 'd'), 0

def prettysize(n):
    pfx, num, magnitude = _prettysize(n)
    return f'{pfx}{num}{_unit[magnitude]}'

def size_color(n):
    _, _, magnitude = _prettysize(n)
    return _unit_color[magnitude]

def prettysize_clr(n, padding=0):
    pfx, num, magnitude = _prettysize(n)
    return Colorizer().t(f'{pfx}{num}{_unit[magnitude]}'.rjust(padding), fg=_unit_color[magnitude])

RX_COMMA = re.compile(r'(\d\d\d)(?=\d)')
def addcomma(n):
    n, sep, d = str(n).partition('.')
    return RX_COMMA.sub(r'\1,', n[::-1])[::-1] + sep + d

mul = dict(k=1000, m=1000000, g=1000000000, t=1000000000000)
def parsesize(s):
    if s is None:
        return None

    s = s.lower().replace(',', '')
    m = mul.get(s[-1:])
    if m:
        return int(float(s[:-1]) * m)
    return int(s)

def read_file(p, default=None):
    try:
        r = p.read_text().strip()
        return r or default
    except FileNotFoundError:
        return default

def write_file(p, data=None):
    tmpf = p.with_name(p.name + '~')
    with tmpf.open('w') as fp:
        fp.write(data + '\n')
        fp.flush()
        os.fsync(fp.fileno())
    tmpf.replace(p)

def try_unlink(p):
    try:
        os.unlink(p)
        return True
    except OSError:
        return False

def check_dir(f):
    f.parent.mkdir(parents=True, exist_ok=True)

def compact_json(data):
    return json.dumps(data, separators=(',',':'))

def canonical_path(path, tslash=False):
    if path is None:
        return None

    # Collapse all multiple slashes into one
    path = re.sub(r'[\\/]+', '/', path)

    # Remove extra slashes from the beginning
    path = '/' + (path.lstrip('/') if tslash is None else path.strip('/'))
    if tslash and path != '/':
        path += '/'
    return path

##########################################################################################
# Functions exported to SQLite
##########################################################################################

SQL_FUNCS = {}
def sqlf(f):
    def ret(*args):
        try:
            return f(*args)
        except Exception:
            traceback.print_exc()
            raise
    SQL_FUNCS[f.__name__] = ret
    return f

@sqlf
def disk_in_nexus(nexus, nexus_index):
    if nexus_index < 0 or nexus_index >= len(nexus):
        return False
    return nexus[nexus_index] == '1'

@sqlf
def nexus_with_disk(nexus, nexus_index):
    if nexus_index >= len(nexus):
        return nexus + '0' * (nexus_index - len(nexus)) + '1'
    return nexus[:nexus_index] + '1' + nexus[nexus_index + 1:]

@sqlf
def nexus_without_disk(nexus, nexus_index):
    if nexus_index >= len(nexus):
        return nexus
    return (nexus[:nexus_index] + '0' + nexus[nexus_index + 1:]).rstrip('0')

@sqlf
def nexus_level(nexus):
    return nexus.count('1')

@sqlf
def new_uuid():
    return str(uuid.uuid4())

@sqlf
def canonical_uuid(val):
    return str(uuid.UUID(val))

def prefix_upper(prefix):
    return prefix[:-1] + chr(1 + ord(prefix[-1:]))

def prefix_sql(column, prefix):
    return f'({column} >= ? AND {column} < ?)', (prefix, prefix_upper(prefix))

def escape_like(text):
    return text.replace('\\', r'\\').replace('%', r'\%').replace('_', r'\_')

def hash_relpath(hash):
    return f'distbackup-objects/{hash[:3]}/{hash}'

def pattern_to_regex(pat):
    if not pat:
        return None, None

    pat = canonical_path(pat, None)

    prefix = None
    m = re.match(r'^([^\*\.\[\?]+)', pat)
    if m:
        prefix = m.group(1)

    rxpattern = fnmatch.translate(pat)

    # fnmatch.translate tacks on \Z to match the end of the string
    if rxpattern.endswith(r'\Z'):
        rxpattern = rxpattern[:-2]

    # Match either an exact file, or any file in a subdirectory
    if pat.endswith('/'):
        rxpattern += r'(.*)?\Z'
    else:
        rxpattern += r'(/.*)?\Z'

    return re.compile(rxpattern, re.S), prefix

def round_blocks(s):
    return 4096 + (s + 4095) & ~4095

def int_mtime(m):
    return int(m * 1000)

def restore_mtime(m):
    return m / 1000.0



import random
from typing import Optional
from dataclasses import dataclass, field
from .nexus import Nexus

@dataclass(order=True)
class ObjectCopyEntry:
    hash: str = field(compare=False)
    size: int = field(compare=False)
    blocksize: int = field(compare=False)
    size_order: int = field(compare=False)
    groupkey: tuple[int, int] = field(default=(0, 0), compare=False)
    maxcopies: Optional[int] = field(default=None, compare=False)
    refs: int = field(default=0, compare=False)
    nexusfrom: Optional[Nexus] = field(compare=False, default=None)
    nexusto: Optional[Nexus] = field(compare=False, default=None)
    path: str = field(default='', compare=False)
    modtime: int = field(default=0, compare=False)
    imbalance_change: float = field(default=0.0, compare=True)

@dataclass(order=True)
class ObjectGroup:
    groupkey: tuple[int, int]
    objects: list[ObjectCopyEntry] = field(default_factory=list, compare=False)
    saved_objects: Optional[list] = field(default=None, compare=False)
    totalsize: int = field(default=0, compare=False)
    sorted: bool = field(default=False, compare=False)

    def commit(self):
        self.saved_objects = None

    def rollback(self):
        if self.saved_objects is not None:
            self.objects = self.saved_objects

    def add_entry(self, obj):
        self.objects.append(obj)
        self.totalsize += obj.blocksize

    def pop_entry(self, max_size=None):
        '''Calculates the imbalance change for all objects in the group, sorts them with
        the best balance improvements at the front, then chooses a random entry from the
        first third of the list, weighted toward earlier entries, removes, and returns
        it.'''

        # Save the previous list of objects in case we have to back out changes
        if self.saved_objects is None:
            self.saved_objects = list(self.objects)

        while True:
            # Generate a uniform random number between 0.0 and 1.0, then cube it to weight it lower.
            rval = random.random()
            pos = int((rval * rval * rval) * len(self.objects) / 3)
            item = self.objects[pos]
            del self.objects[pos]
            self.totalsize -= item.blocksize
            if max_size is None or item.blocksize <= max_size:
                return item

    def sort(self, levels):
        '''Recalculate the imbalance change for all objects in the group'''
        for obj in self.objects:
            obj.imbalance_change = levels.calc_imbalance_change(obj.nexusfrom, obj.nexusto, obj.size_order)
        self.objects.sort()
        self.sorted = True

class ObjectGroupList(list):
    def __init__(self, ondisk):
        super().__init__()
        self.ondisk = ondisk

    def commit(self):
        for grp in self:
            grp.commit()

    def rollback(self):
        for grp in self:
            grp.rollback()
